@@ -7,15 +7,14 @@
 #
 # See the file "license.terms" for information on usage and redistribution
 # of this file, and for a DISCLAIMER OF ALL WARRANTIES.
-# 
-# RCS: @(#) $Id: math.tcl,v 1.20 2005/10/06 05:16:37 andreas_kupries Exp $
 
 package require Tcl 8.2		;# uses [lindex $l end-$integer]
 
+# @mdgen OWNER: tclIndex
+# @mdgen OWNER: misc.tcl
+# @mdgen OWNER: combinatorics.tcl
+
 namespace eval ::math {
-
-    variable version 1.2.3
-
     # misc.tcl
 
     namespace export	cov		fibonacci	integrate
@@ -31,10 +30,15 @@ namespace eval ::math {
 
     # Set up for auto-loading
 
-    variable home [file join [pwd] [file dirname [info script]]]
-    if {[lsearch -exact $::auto_path $home] == -1} {
-	lappend ::auto_path $home
+    if { ![interp issafe {}]} {
+	variable home [file join [pwd] [file dirname [info script]]]
+	if {[lsearch -exact $::auto_path $home] == -1} {
+	    lappend ::auto_path $home
+	}
+    } else {
+	source [file join [file dirname [info script]] misc.tcl]
+	source [file join [file dirname [info script]] combinatorics.tcl]
     }
 
-    package provide [namespace tail [namespace current]] $version
+    package provide [namespace tail [namespace current]] 1.2.5
 }
